"""Tests for Flex Stacker Engine Core."""

import inspect
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.errors.exceptions import CommandPreconditionViolated
from opentrons_shared_data.labware.labware_definition import LabwareDefinition2

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import FlexStacker
from opentrons.hardware_control.modules.types import (
    ModuleType,
)
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, OFF_DECK
from opentrons.protocol_api.core.engine import load_labware_params
from opentrons.protocol_api.core.engine.labware import LabwareCore
from opentrons.protocol_api.core.engine.module_core import FlexStackerCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors.exceptions import (
    FlexStackerLabwarePoolNotYetDefinedError,
)
from opentrons.protocol_engine.types import (
    OverlapOffset,
    StackerFillEmptyStrategy,
    StackerStoredLabwareGroup,
)

SyncFlexStackerHardware = SynchronousAdapter[FlexStacker]


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SyncFlexStackerHardware:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SyncFlexStackerHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture(autouse=True)
def patch_mock_load_labware_params(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out load_labware_params.py functions."""
    for name, func in inspect.getmembers(load_labware_params, inspect.isfunction):
        monkeypatch.setattr(load_labware_params, name, decoy.mock(func=func))


@pytest.fixture
def subject(
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncFlexStackerHardware,
    mock_protocol_core: ProtocolCore,
) -> FlexStackerCore:
    """Get a Flex Stacker Core test subject."""
    return FlexStackerCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncFlexStackerHardware,
    mock_protocol_core: ProtocolCore,
) -> None:
    """It should be able to create a Flex Stacker module core."""
    result = FlexStackerCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )

    assert result.module_id == "1234"
    assert result.MODULE_TYPE == ModuleType.FLEX_STACKER


def test_set_stored_labware_all_elements(
    decoy: Decoy, mock_engine_client: EngineClient, subject: FlexStackerCore
) -> None:
    """It should gather labware into objects appropriately when specifying all of lid, labware, and adapter."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return(sentinel.custom_labware_load_params)
    decoy.when(
        load_labware_params.resolve(
            "main-name",
            "main-namespace",
            1,
            sentinel.custom_labware_load_params,
            MAX_SUPPORTED_VERSION,
        )
    ).then_return(("main-namespace-verified", 10))
    decoy.when(
        load_labware_params.resolve(
            "adapter-name",
            "adapter-namespace",
            2,
            sentinel.custom_labware_load_params,
            MAX_SUPPORTED_VERSION,
        )
    ).then_return(("adapter-namespace-verified", 20))
    decoy.when(
        load_labware_params.resolve(
            "lid-name",
            "lid-namespace",
            3,
            sentinel.custom_labware_load_params,
            MAX_SUPPORTED_VERSION,
        )
    ).then_return(("lid-namespace-verified", 30))

    subject.set_stored_labware(
        main_load_name="main-name",
        main_namespace="main-namespace",
        main_version=1,
        lid_load_name="lid-name",
        lid_namespace="lid-namespace",
        lid_version=3,
        adapter_load_name="adapter-name",
        adapter_namespace="adapter-namespace",
        adapter_version=2,
        count=5,
        stacking_offset_z=2.0,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId="1234",
                initialCount=5,
                primaryLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="main-name",
                    namespace="main-namespace-verified",
                    version=10,
                ),
                lidLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="lid-name", namespace="lid-namespace-verified", version=30
                ),
                adapterLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="adapter-name",
                    namespace="adapter-namespace-verified",
                    version=20,
                ),
                poolOverlapOverride=2.0,
            )
        )
    )


def test_set_stored_labware_only_checks_load_name_for_lid_and_adapter_valid(
    decoy: Decoy, mock_engine_client: EngineClient, subject: FlexStackerCore
) -> None:
    """It should specify lid and adapter if and only if their load names are specified."""
    decoy.when(
        mock_engine_client.state.labware.find_custom_labware_load_params()
    ).then_return(sentinel.custom_labware_load_params)
    decoy.when(
        load_labware_params.resolve(
            "main-name",
            "main-namespace",
            1,
            sentinel.custom_labware_load_params,
            MAX_SUPPORTED_VERSION,
        )
    ).then_return(("main-namespace-verified", 10))

    subject.set_stored_labware(
        main_load_name="main-name",
        main_namespace="main-namespace",
        main_version=1,
        lid_load_name=None,
        lid_namespace="lid-namespace",
        lid_version=3,
        adapter_load_name=None,
        adapter_namespace="adapter-namespace",
        adapter_version=2,
        count=5,
    )
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId="1234",
                initialCount=5,
                primaryLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="main-name",
                    namespace="main-namespace-verified",
                    version=10,
                ),
                lidLabware=None,
                adapterLabware=None,
            )
        )
    )


def test_get_max_storable_labware_from_empty_list(subject: FlexStackerCore) -> None:
    """It should handle an empty list of labware."""
    assert subject.get_max_storable_labware_from_list([]) == []


def test_get_current_storable_labware_from_empty_list(subject: FlexStackerCore) -> None:
    """It should handle an empty list of labware."""
    assert subject.get_current_storable_labware_from_list([]) == []


@pytest.mark.parametrize("overlap_override", [None, sentinel.override_value])
def test_get_max_storable_labware_from_list_configured(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
    overlap_override: float | None,
) -> None:
    """It should use the substate's max pool count."""
    cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(3)
    assert (
        subject.get_max_storable_labware_from_list(cores, overlap_override) == cores[:3]
    )
    # If an overlap override is specified, it should be passed to the engine to make
    # sure the value matches the substate's value
    decoy.verify(
        mock_engine_client.state.modules.validate_stacker_overlap_offset(
            "1234", sentinel.override_value
        ),
        times=1 if overlap_override is not None else 0,
    )


def test_get_max_storable_labware_from_list_unconfigured_primary_only(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should marshal labware and ask the engine."""
    # the stacker has not been configured, so it should return None
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(None)

    cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    decoy.when(mock_protocol_core.get_labware_location(cores[0])).then_return(OFF_DECK)
    decoy.when(mock_protocol_core.get_labware_on_labware(cores[0])).then_return(None)
    decoy.when(cores[0].labware_id).then_return("core-0")
    primary_def = decoy.mock(cls=LabwareDefinition2)
    decoy.when(cores[0].get_engine_definition()).then_return(primary_def)
    decoy.when(primary_def.parameters.loadName).then_return("mock-load-name")
    decoy.when(
        mock_engine_client.state.labware.stacker_labware_pool_to_ordered_list(
            primary_labware_definition=primary_def,
            lid_labware_definition=None,
            adapter_labware_definition=None,
        )
    ).then_return([primary_def])
    decoy.when(
        mock_engine_client.state.geometry.get_height_of_labware_stack([primary_def])
    ).then_return(10)
    decoy.when(
        mock_engine_client.state.labware.get_stacker_labware_overlap_offset(
            [primary_def]
        )
    ).then_return(OverlapOffset(x=0, y=0, z=2))
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count_by_height("1234", 10, 2)
    ).then_return(3)
    assert subject.get_max_storable_labware_from_list(cores) == cores[:3]


def test_get_max_storable_labware_from_list_unconfigured_no_lid(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should marshal labware and ask the engine."""
    # the stacker has not been configured, so it should return None
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(None)

    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    adapter_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    decoy.when(mock_protocol_core.get_labware_location(primary_cores[0])).then_return(
        adapter_cores[0]
    )
    decoy.when(mock_protocol_core.get_labware_on_labware(primary_cores[0])).then_return(
        None
    )
    decoy.when(primary_cores[0].labware_id).then_return("core-0")
    decoy.when(adapter_cores[0].labware_id).then_return("adapter-0")

    primary_def = decoy.mock(cls=LabwareDefinition2)
    adapter_def = decoy.mock(cls=LabwareDefinition2)
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_def)
    decoy.when(adapter_cores[0].get_engine_definition()).then_return(adapter_def)
    decoy.when(adapter_def.parameters.loadName).then_return("mock-adapter-name")
    decoy.when(
        mock_engine_client.state.labware.stacker_labware_pool_to_ordered_list(
            primary_labware_definition=primary_def,
            adapter_labware_definition=adapter_def,
            lid_labware_definition=None,
        )
    ).then_return([primary_def, adapter_def])
    decoy.when(
        mock_engine_client.state.geometry.get_height_of_labware_stack(
            [primary_def, adapter_def]
        )
    ).then_return(16)
    decoy.when(
        mock_engine_client.state.labware.get_stacker_labware_overlap_offset(
            [primary_def, adapter_def]
        )
    ).then_return(OverlapOffset(x=0, y=0, z=3.0))
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count_by_height(
            "1234", 16, 3.0
        )
    ).then_return(3)
    assert (
        subject.get_max_storable_labware_from_list(primary_cores) == primary_cores[:3]
    )


def test_get_max_storable_labware_from_list_unconfigured_no_adapter(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should marshal labware and ask the engine."""
    # the stacker has not been configured, so it should return None
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(None)

    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    lid_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    decoy.when(mock_protocol_core.get_labware_location(primary_cores[0])).then_return(
        OFF_DECK
    )
    decoy.when(mock_protocol_core.get_labware_on_labware(primary_cores[0])).then_return(
        lid_cores[0]
    )
    decoy.when(primary_cores[0].labware_id).then_return("core-0")
    decoy.when(lid_cores[0].labware_id).then_return("lid-0")

    primary_def = decoy.mock(cls=LabwareDefinition2)
    lid_def = decoy.mock(cls=LabwareDefinition2)
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_def)
    decoy.when(lid_cores[0].get_engine_definition()).then_return(lid_def)
    decoy.when(primary_def.parameters.loadName).then_return("mock-primary-name")
    decoy.when(
        mock_engine_client.state.labware.stacker_labware_pool_to_ordered_list(
            primary_labware_definition=primary_def,
            lid_labware_definition=lid_def,
            adapter_labware_definition=None,
        )
    ).then_return([lid_def, primary_def])
    decoy.when(
        mock_engine_client.state.geometry.get_height_of_labware_stack(
            [lid_def, primary_def]
        )
    ).then_return(11)
    decoy.when(
        mock_engine_client.state.labware.get_stacker_labware_overlap_offset(
            [lid_def, primary_def]
        )
    ).then_return(OverlapOffset(x=0, y=0, z=1.0))
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count_by_height(
            "1234", 11, 1.0
        )
    ).then_return(3)
    assert (
        subject.get_max_storable_labware_from_list(primary_cores) == primary_cores[:3]
    )


def test_get_max_storable_labware_from_list_unconfigured_full_group(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should marshal labware and ask the engine."""
    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    adapter_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    lid_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    decoy.when(mock_protocol_core.get_labware_location(primary_cores[0])).then_return(
        adapter_cores[0]
    )
    decoy.when(mock_protocol_core.get_labware_on_labware(primary_cores[0])).then_return(
        lid_cores[0]
    )
    decoy.when(primary_cores[0].labware_id).then_return("core-0")
    decoy.when(adapter_cores[0].labware_id).then_return("adapter-0")
    decoy.when(lid_cores[0].labware_id).then_return("lid-0")

    primary_def = decoy.mock(cls=LabwareDefinition2)
    adapter_def = decoy.mock(cls=LabwareDefinition2)
    lid_def = decoy.mock(cls=LabwareDefinition2)
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_def)
    decoy.when(adapter_cores[0].get_engine_definition()).then_return(adapter_def)
    decoy.when(lid_cores[0].get_engine_definition()).then_return(lid_def)
    decoy.when(adapter_def.parameters.loadName).then_return("mock-adapter-name")
    decoy.when(
        mock_engine_client.state.labware.stacker_labware_pool_to_ordered_list(
            primary_labware_definition=primary_def,
            lid_labware_definition=lid_def,
            adapter_labware_definition=adapter_def,
        )
    ).then_return([lid_def, primary_def, adapter_def])
    decoy.when(
        mock_engine_client.state.geometry.get_height_of_labware_stack(
            [lid_def, primary_def, adapter_def]
        )
    ).then_return(20)
    decoy.when(
        mock_engine_client.state.labware.get_stacker_labware_overlap_offset(
            [lid_def, primary_def, adapter_def]
        )
    ).then_return(OverlapOffset(x=0, y=0, z=2))
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count_by_height("1234", 20, 2)
    ).then_return(3)
    assert (
        subject.get_max_storable_labware_from_list(primary_cores) == primary_cores[:3]
    )


def test_get_current_storable_labware_from_list_primary_happypath(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should marshal labware and ask the engine."""
    cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(5)
    decoy.when(
        mock_engine_client.state.modules.stacker_contained_labware("1234")
    ).then_return([sentinel.lw1, sentinel.lw2])
    assert subject.get_current_storable_labware_from_list(cores) == cores[:3]


def test_get_current_storable_labware_happypath(
    decoy: Decoy, mock_engine_client: EngineClient, subject: FlexStackerCore
) -> None:
    """It should get storable labware when everything is configured."""
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(10)
    decoy.when(
        mock_engine_client.state.modules.stacker_contained_labware("1234")
    ).then_return([sentinel.lw1, sentinel.lw2])
    assert subject.get_current_storable_labware() == 8


def test_get_max_storable_labware_happypath(
    decoy: Decoy, mock_engine_client: EngineClient, subject: FlexStackerCore
) -> None:
    """It should get max labware when everything is configured."""
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(10)
    assert subject.get_max_storable_labware() == 10


def test_get_max_storable_labware_raises_if_not_configured(
    decoy: Decoy, mock_engine_client: EngineClient, subject: FlexStackerCore
) -> None:
    """It should raise an exception if the stacker isn't configured."""
    decoy.when(
        mock_engine_client.state.modules.stacker_max_pool_count("1234")
    ).then_return(None)
    with pytest.raises(FlexStackerLabwarePoolNotYetDefinedError):
        subject.get_max_storable_labware()


def _prime_definition(decoy: Decoy, name: str) -> LabwareDefinition2:
    def_object = decoy.mock(cls=LabwareDefinition2)
    decoy.when(def_object.parameters.loadName).then_return(f"{name}-loadname")
    decoy.when(def_object.namespace).then_return(f"{name}-namespace")
    decoy.when(def_object.version).then_return(1)
    return def_object


def test_set_stored_labware_items_happypath_primary_only(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should set stored labware if all is well."""
    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, primary in enumerate(primary_cores):
        decoy.when(mock_protocol_core.get_labware_location(primary)).then_return(
            OFF_DECK
        )
        decoy.when(mock_protocol_core.get_labware_on_labware(primary)).then_return(None)
        decoy.when(primary.labware_id).then_return(f"primary-{idx}")
    primary_definition = _prime_definition(decoy, "primary")
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_definition)
    subject.set_stored_labware_items(primary_cores, None)
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId="1234",
                initialCount=None,
                initialStoredLabware=[
                    StackerStoredLabwareGroup(
                        primaryLabwareId=f"primary-{idx}",
                        adapterLabwareId=None,
                        lidLabwareId=None,
                    )
                    for idx in range(5)
                ],
                primaryLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="primary-loadname",
                    namespace="primary-namespace",
                    version=1,
                ),
                adapterLabware=None,
                lidLabware=None,
            )
        )
    )


def test_set_stored_labware_items_happypath_no_adapter(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should set stored labware if all is well."""
    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    lid_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, (primary, lid) in enumerate(zip(primary_cores, lid_cores)):
        decoy.when(mock_protocol_core.get_labware_location(primary)).then_return(
            OFF_DECK
        )
        decoy.when(mock_protocol_core.get_labware_on_labware(primary)).then_return(lid)
        decoy.when(primary.labware_id).then_return(f"primary-{idx}")
        decoy.when(lid.labware_id).then_return(f"lid-{idx}")
    primary_definition = _prime_definition(decoy, "primary")
    lid_definition = _prime_definition(decoy, "lid")
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_definition)
    decoy.when(lid_cores[0].get_engine_definition()).then_return(lid_definition)
    subject.set_stored_labware_items(primary_cores, None)
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId="1234",
                initialCount=None,
                initialStoredLabware=[
                    StackerStoredLabwareGroup(
                        primaryLabwareId=f"primary-{idx}",
                        adapterLabwareId=None,
                        lidLabwareId=f"lid-{idx}",
                    )
                    for idx in range(5)
                ],
                primaryLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="primary-loadname",
                    namespace="primary-namespace",
                    version=1,
                ),
                adapterLabware=None,
                lidLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="lid-loadname", namespace="lid-namespace", version=1
                ),
            )
        )
    )


def test_set_stored_labware_items_happypath_no_lid(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should set stored labware if all is well."""
    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    adapter_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, (primary, adapter) in enumerate(zip(primary_cores, adapter_cores)):
        decoy.when(mock_protocol_core.get_labware_location(primary)).then_return(
            adapter
        )
        decoy.when(mock_protocol_core.get_labware_on_labware(primary)).then_return(None)
        decoy.when(primary.labware_id).then_return(f"primary-{idx}")
        decoy.when(adapter.labware_id).then_return(f"adapter-{idx}")
    primary_definition = _prime_definition(decoy, "primary")
    adapter_definition = _prime_definition(decoy, "adapter")
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_definition)
    decoy.when(adapter_cores[0].get_engine_definition()).then_return(adapter_definition)
    subject.set_stored_labware_items(primary_cores, None)
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId="1234",
                initialCount=None,
                initialStoredLabware=[
                    StackerStoredLabwareGroup(
                        primaryLabwareId=f"primary-{idx}",
                        adapterLabwareId=f"adapter-{idx}",
                        lidLabwareId=None,
                    )
                    for idx in range(5)
                ],
                primaryLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="primary-loadname",
                    namespace="primary-namespace",
                    version=1,
                ),
                adapterLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="adapter-loadname",
                    namespace="adapter-namespace",
                    version=1,
                ),
                lidLabware=None,
            )
        )
    )


def test_set_stored_labware_items_happypath_all_elements(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should set stored labware if all is well."""
    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    adapter_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    lid_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, (primary, adapter, lid) in enumerate(
        zip(primary_cores, adapter_cores, lid_cores)
    ):
        decoy.when(mock_protocol_core.get_labware_location(primary)).then_return(
            adapter
        )
        decoy.when(mock_protocol_core.get_labware_on_labware(primary)).then_return(lid)
        decoy.when(primary.labware_id).then_return(f"primary-{idx}")
        decoy.when(adapter.labware_id).then_return(f"adapter-{idx}")
        decoy.when(lid.labware_id).then_return(f"lid-{idx}")
    primary_definition = _prime_definition(decoy, "primary")
    adapter_definition = _prime_definition(decoy, "adapter")
    lid_definition = _prime_definition(decoy, "lid")
    decoy.when(primary_cores[0].get_engine_definition()).then_return(primary_definition)
    decoy.when(adapter_cores[0].get_engine_definition()).then_return(adapter_definition)
    decoy.when(lid_cores[0].get_engine_definition()).then_return(lid_definition)
    subject.set_stored_labware_items(primary_cores, None)
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId="1234",
                initialCount=None,
                initialStoredLabware=[
                    StackerStoredLabwareGroup(
                        primaryLabwareId=f"primary-{idx}",
                        adapterLabwareId=f"adapter-{idx}",
                        lidLabwareId=f"lid-{idx}",
                    )
                    for idx in range(5)
                ],
                primaryLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="primary-loadname",
                    namespace="primary-namespace",
                    version=1,
                ),
                adapterLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="adapter-loadname",
                    namespace="adapter-namespace",
                    version=1,
                ),
                lidLabware=cmd.flex_stacker.StackerStoredLabwareDetails(
                    loadName="lid-loadname", namespace="lid-namespace", version=1
                ),
            )
        )
    )


def test_set_stored_items_labware_fails_if_empty(
    subject: FlexStackerCore,
) -> None:
    """It should fail if given an empty list."""
    with pytest.raises(CommandPreconditionViolated):
        subject.set_stored_labware_items([], None)


def test_fill_items_happypath(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_protocol_core: ProtocolCore,
    subject: FlexStackerCore,
) -> None:
    """It should call fill items properly."""
    primary_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    adapter_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    lid_cores = [decoy.mock(cls=LabwareCore) for _ in range(5)]
    for idx, (primary, adapter, lid) in enumerate(
        zip(primary_cores, adapter_cores, lid_cores)
    ):
        decoy.when(mock_protocol_core.get_labware_location(primary)).then_return(
            adapter
        )
        decoy.when(mock_protocol_core.get_labware_on_labware(primary)).then_return(lid)
        decoy.when(primary.labware_id).then_return(f"primary-{idx}")
        decoy.when(adapter.labware_id).then_return(f"adapter-{idx}")
        decoy.when(lid.labware_id).then_return(f"lid-{idx}")
    subject.fill_items(primary_cores, "hello")
    decoy.verify(
        mock_engine_client.execute_command(
            cmd.flex_stacker.FillParams(
                moduleId="1234",
                count=None,
                strategy=StackerFillEmptyStrategy.MANUAL_WITH_PAUSE,
                labwareToStore=[
                    StackerStoredLabwareGroup(
                        primaryLabwareId=f"primary-{idx}",
                        adapterLabwareId=f"adapter-{idx}",
                        lidLabwareId=f"lid-{idx}",
                    )
                    for idx in range(5)
                ],
                message="hello",
            )
        )
    )



"""Labware origin math errors."""

from typing import Any, Dict, Optional, Sequence

from opentrons_shared_data.errors import ErrorCodes
from opentrons_shared_data.errors.exceptions import EnumeratedError

from opentrons.protocol_engine.errors import ProtocolEngineError


class LabwareLocatingFeatureError(ProtocolEngineError):
    """Base class for errors related to labware locating features."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareLocatingFeatureError."""
        super().__init__(
            ErrorCodes.LABWARE_LOCATING_FEATURE_ERROR, message, details, wrapping
        )


class MissingLocatingFeatureError(LabwareLocatingFeatureError):
    """Raised when a labware definition is missing a required locating feature."""

    def __init__(
        self,
        labware_name: str,
        required_feature: str,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MissingLocatingFeatureError."""
        if message is None:
            message = f"Expected {labware_name} to have {required_feature} feature"

        if details is None:
            details = {
                "labware_name": labware_name,
                "required_feature": required_feature,
            }

        super().__init__(message, details, wrapping)


class InvalidLabwarePlacementError(LabwareLocatingFeatureError):
    """Raised when a labware cannot be placed in the specified location due to locating feature constraints."""

    def __init__(
        self,
        feature_name: str,
        invalid_placement: str,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidLabwarePlacementError."""
        if message is None:
            message = f"{feature_name} feature does not support placement: {invalid_placement}"

        if details is None:
            details = {
                "feature_name": feature_name,
                "invalid_placement": invalid_placement,
            }

        super().__init__(message, details, wrapping)


class IncompatibleLocatingFeatureError(LabwareLocatingFeatureError):
    """Raised when parent and child labware have incompatible locating features."""

    def __init__(
        self,
        parent_feature: str,
        child_feature: str,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an IncompatibleLocatingFeatureError."""
        if message is None:
            message = f"Incompatible labware features: parent {parent_feature}, child {child_feature}"

        if details is None:
            details = {
                "parent_feature": parent_feature,
                "child_feature": child_feature,
            }

        super().__init__(message, details, wrapping)



"""Tests for CommandStore.

DEPRECATED: Testing CommandStore independently of CommandView is no longer helpful.
Try to add new tests to test_command_state.py, where they can be tested together,
treating CommandState as a private implementation detail.
"""

from datetime import datetime
from typing import Any

import pytest
from decoy import matchers

from opentrons_shared_data.errors import ErrorCodes

from .command_fixtures import create_succeeded_command
from opentrons.ordered_set import OrderedSet
from opentrons.protocol_engine import commands, errors
from opentrons.protocol_engine.actions import (
    FinishAction,
    FinishErrorDetails,
    HardwareStoppedAction,
    PauseAction,
    PauseSource,
    PlayAction,
    QueueCommandAction,
    StopAction,
    SucceedCommandAction,
)
from opentrons.protocol_engine.actions.actions import RunCommandAction
from opentrons.protocol_engine.state.command_history import CommandEntry, CommandHistory
from opentrons.protocol_engine.state.commands import (
    CommandState,
    CommandStore,
    QueueStatus,
    RunResult,
)
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.types import DeckType


def _make_config(block_on_door_open: bool = False) -> Config:
    return Config(
        block_on_door_open=block_on_door_open,
        # Choice of robot and deck type is arbitrary.
        robot_type="OT-2 Standard",
        deck_type=DeckType.OT2_STANDARD,
    )


def _placeholder_error_recovery_policy(*args: object, **kwargs: object) -> Any:
    """A placeholder `ErrorRecoveryPolicy` for tests that don't care about it.

    That should be all the tests in this file, since error recovery was added
    after this file was deprecated.
    """
    raise NotImplementedError()


def test_command_queue_and_unqueue() -> None:
    """It should queue on QueueCommandAction and dequeue on RunCommandAction."""
    queue_1 = QueueCommandAction(
        request=commands.WaitForResumeCreate(params=commands.WaitForResumeParams()),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    queue_2 = QueueCommandAction(
        request=commands.WaitForResumeCreate(params=commands.WaitForResumeParams()),
        request_hash=None,
        created_at=datetime(year=2022, month=2, day=2),
        command_id="command-id-2",
    )
    run_1 = RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2021, month=1, day=1),
    )
    run_2 = RunCommandAction(
        command_id="command-id-2",
        started_at=datetime(year=2022, month=2, day=2),
    )
    succeed_2 = SucceedCommandAction(
        command=create_succeeded_command(command_id="command-id-2"),
    )

    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(queue_1)
    assert subject.state.command_history.get_queue_ids() == OrderedSet(["command-id-1"])

    subject.handle_action(queue_2)
    assert subject.state.command_history.get_queue_ids() == OrderedSet(
        ["command-id-1", "command-id-2"]
    )

    subject.handle_action(run_2)
    assert subject.state.command_history.get_queue_ids() == OrderedSet(["command-id-1"])

    subject.handle_action(succeed_2)
    subject.handle_action(run_1)
    assert subject.state.command_history.get_queue_ids() == OrderedSet()


def test_setup_command_queue_and_unqueue() -> None:
    """It should queue and dequeue on setup commands."""
    queue_1 = QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    queue_2 = QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
        ),
        request_hash=None,
        created_at=datetime(year=2022, month=2, day=2),
        command_id="command-id-2",
    )
    run_1 = RunCommandAction(
        command_id="command-id-1", started_at=datetime(year=2021, month=1, day=1)
    )
    run_2 = RunCommandAction(
        command_id="command-id-2", started_at=datetime(year=2022, month=2, day=2)
    )
    succeed_2 = SucceedCommandAction(
        command=create_succeeded_command(command_id="command-id-2"),
    )

    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(queue_1)
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet(
        ["command-id-1"]
    )

    subject.handle_action(queue_2)
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet(
        ["command-id-1", "command-id-2"]
    )

    subject.handle_action(run_2)
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet(
        ["command-id-1"]
    )

    subject.handle_action(succeed_2)
    subject.handle_action(run_1)
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_setup_queue_action_updates_command_intent() -> None:
    """It should update command source correctly."""
    queue_cmd = QueueCommandAction(
        request=commands.WaitForResumeCreate(
            params=commands.WaitForResumeParams(),
            intent=commands.CommandIntent.SETUP,
            key="command-key-1",
        ),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )

    expected_pause_cmd = commands.WaitForResume(
        id="command-id-1",
        key="command-key-1",
        createdAt=datetime(year=2021, month=1, day=1),
        params=commands.WaitForResumeParams(),
        status=commands.CommandStatus.QUEUED,
        intent=commands.CommandIntent.SETUP,
    )

    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(queue_cmd)
    assert subject.state.command_history.get("command-id-1") == CommandEntry(
        index=0, command=expected_pause_cmd
    )


def test_running_command_id() -> None:
    """It should update the running command ID through a command's lifecycle."""
    queue = QueueCommandAction(
        request=commands.WaitForResumeCreate(params=commands.WaitForResumeParams()),
        request_hash=None,
        created_at=datetime(year=2021, month=1, day=1),
        command_id="command-id-1",
    )
    run = RunCommandAction(
        command_id="command-id-1",
        started_at=datetime(year=2021, month=1, day=1),
    )
    succeed = SucceedCommandAction(
        command=create_succeeded_command(command_id="command-id-1"),
    )

    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(queue)
    assert subject.state.command_history.get_running_command() is None

    subject.handle_action(run)
    running_command = subject.state.command_history.get_running_command()
    assert running_command is not None
    assert running_command.command.id == "command-id-1"

    subject.handle_action(succeed)
    assert subject.state.command_history.get_running_command() is None


def test_command_store_keeps_commands_in_queue_order() -> None:
    """It should keep commands in the order they were originally enqueued."""
    command_create_1_non_setup = commands.CommentCreate(
        params=commands.CommentParams(message="hello world"),
    )
    command_create_2_setup = commands.CommentCreate(
        params=commands.CommentParams(message="hello world"),
        intent=commands.CommandIntent.SETUP,
    )
    command_create_3_non_setup = commands.CommentCreate(
        params=commands.CommentParams(message="hello world"),
    )

    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(
        QueueCommandAction(
            "command-id-1",
            created_at=datetime(year=2021, month=1, day=1),
            request=command_create_1_non_setup,
            request_hash=None,
        )
    )
    assert subject.state.command_history.get_all_ids() == ["command-id-1"]

    subject.handle_action(
        QueueCommandAction(
            "command-id-2",
            created_at=datetime(year=2021, month=1, day=1),
            request=command_create_2_setup,
            request_hash=None,
        )
    )
    assert subject.state.command_history.get_all_ids() == [
        "command-id-1",
        "command-id-2",
    ]

    subject.handle_action(
        QueueCommandAction(
            "command-id-3",
            created_at=datetime(year=2021, month=1, day=1),
            request=command_create_3_non_setup,
            request_hash=None,
        )
    )
    assert subject.state.command_history.get_all_ids() == [
        "command-id-1",
        "command-id-2",
        "command-id-3",
    ]

    # Running and completing commands shouldn't affect the command order.
    subject.handle_action(
        RunCommandAction(
            command_id="command-id-2", started_at=datetime(year=2021, month=1, day=1)
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=create_succeeded_command(
                command_id="command-id-2",
            ),
        )
    )
    assert subject.state.command_history.get_all_ids() == [
        "command-id-1",
        "command-id-2",
        "command-id-3",
    ]


@pytest.mark.parametrize("pause_source", PauseSource)
def test_command_store_handles_pause_action(pause_source: PauseSource) -> None:
    """It should clear the running flag on pause."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject.handle_action(PauseAction(source=pause_source))

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=None,
        run_completed_at=None,
        run_started_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )


@pytest.mark.parametrize("pause_source", PauseSource)
def test_command_store_handles_play_action(pause_source: PauseSource) -> None:
    """It should set the running flag on play."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.RUNNING,
        run_result=None,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=datetime(year=2021, month=1, day=1),
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_handles_finish_action() -> None:
    """It should change to a succeeded state with FinishAction."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))
    subject.handle_action(FinishAction())

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.SUCCEEDED,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=datetime(year=2021, month=1, day=1),
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_handles_finish_action_with_stopped() -> None:
    """It should change to a stopped state if FinishAction has set_run_status=False."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))
    subject.handle_action(FinishAction(set_run_status=False))

    assert subject.state.run_result == RunResult.STOPPED


@pytest.mark.parametrize(
    ["from_asynchronous_error", "expected_run_result"],
    [(True, RunResult.FAILED), (False, RunResult.STOPPED)],
)
def test_command_store_handles_stop_action(
    from_asynchronous_error: bool, expected_run_result: RunResult
) -> None:
    """It should mark the engine as non-gracefully stopped on StopAction."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))
    subject.handle_action(StopAction(from_asynchronous_error=from_asynchronous_error))

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=expected_run_result,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=datetime(year=2021, month=1, day=1),
        latest_protocol_command_hash=None,
        stopped_by_async_error=from_asynchronous_error,
        is_stopping_because_of_async_error=from_asynchronous_error,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_handles_stop_action_when_awaiting_recovery() -> None:
    """It should mark the engine as non-gracefully stopped on StopAction."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))

    subject.state.queue_status = QueueStatus.AWAITING_RECOVERY

    subject.handle_action(StopAction())

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.STOPPED,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=datetime(year=2021, month=1, day=1),
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_cannot_restart_after_should_stop() -> None:
    """It should reject a play action after finish."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    subject.handle_action(FinishAction())
    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.SUCCEEDED,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=None,
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_save_started_completed_run_timestamp() -> None:
    """It should save started and completed timestamps."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )
    start_time = datetime(year=2021, month=1, day=1)
    hardware_stopped_time = datetime(year=2022, month=2, day=2)

    subject.handle_action(PlayAction(requested_at=start_time))
    subject.handle_action(
        HardwareStoppedAction(
            completed_at=hardware_stopped_time, finish_error_details=None
        )
    )

    assert subject.state.run_started_at == start_time
    assert subject.state.run_completed_at == hardware_stopped_time


def test_timestamps_are_latched() -> None:
    """It should not change startedAt or completedAt once set."""
    subject = CommandStore(
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
        is_door_open=False,
    )

    play_time_1 = datetime(year=2021, month=1, day=1)
    play_time_2 = datetime(year=2022, month=2, day=2)
    stop_time_1 = datetime(year=2023, month=3, day=3)
    stop_time_2 = datetime(year=2024, month=4, day=4)

    subject.handle_action(PlayAction(requested_at=play_time_1))
    subject.handle_action(PauseAction(source=PauseSource.CLIENT))
    subject.handle_action(PlayAction(requested_at=play_time_2))
    subject.handle_action(
        HardwareStoppedAction(completed_at=stop_time_1, finish_error_details=None)
    )
    subject.handle_action(
        HardwareStoppedAction(completed_at=stop_time_2, finish_error_details=None)
    )

    assert subject.state.run_started_at == play_time_1
    assert subject.state.run_completed_at == stop_time_1


def test_command_store_wraps_unknown_errors() -> None:
    """Fatal errors that are unknown should be wrapped in EnumeratedErrors.

    Fatal errors can come in through FinishActions and HardwareStoppedActions.
    If these are not descendants of EnumeratedError already, they should be
    wrapped in an EnumeratedError before being converted to an ErrorOccurrence.

    The wrapping EnumeratedError should be an UnexpectedProtocolError for errors that happened
    in the main part of the protocol run, or a PythonException for errors that happened elsewhere.
    """
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(
        FinishAction(
            error_details=FinishErrorDetails(
                error=RuntimeError("oh no"),
                error_id="error-id-1",
                created_at=datetime(year=2021, month=1, day=1),
            )
        )
    )

    subject.handle_action(
        HardwareStoppedAction(
            completed_at=datetime(year=2022, month=2, day=2),
            finish_error_details=FinishErrorDetails(
                error=RuntimeError("yikes"),
                error_id="error-id-2",
                created_at=datetime(year=2023, month=3, day=3),
            ),
        )
    )

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.FAILED,
        run_completed_at=datetime(year=2022, month=2, day=2),
        is_door_blocking=False,
        run_error=errors.ErrorOccurrence(
            id="error-id-1",
            createdAt=datetime(year=2021, month=1, day=1),
            # This is wrapped into an UnexpectedProtocolError because it's not
            # enumerated, and it happened in the main part of the run.
            errorType="UnexpectedProtocolError",
            # Unknown errors use the default error code.
            errorCode=ErrorCodes.GENERAL_ERROR.value.code,
            # And it has information about what created it.
            detail="oh no",
            wrappedErrors=[
                errors.ErrorOccurrence(
                    id="error-id-1",
                    createdAt=datetime(year=2021, month=1, day=1),
                    errorType="PythonException",
                    detail="RuntimeError: oh no",
                    errorCode="4000",
                    errorInfo={
                        "class": "RuntimeError",
                        "args": "('oh no',)",
                    },
                    wrappedErrors=[],
                )
            ],
        ),
        finish_error=errors.ErrorOccurrence(
            id="error-id-2",
            createdAt=datetime(year=2023, month=3, day=3),
            # This is wrapped into a PythonException because it's not
            # enumerated, and it happened during the post-run cleanup steps.
            errorType="PythonException",
            # Unknown errors use the default error code.
            errorCode=ErrorCodes.GENERAL_ERROR.value.code,
            # And it has information about what created it.
            detail="RuntimeError: yikes",
            errorInfo={
                "class": "RuntimeError",
                "args": "('yikes',)",
            },
        ),
        run_started_at=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_preserves_enumerated_errors() -> None:
    """If an error is derived from EnumeratedError, it should be stored as-is."""

    class MyCustomError(errors.ProtocolEngineError):
        def __init__(self, message: str) -> None:
            super().__init__(ErrorCodes.PIPETTE_NOT_PRESENT, message)

    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(
        FinishAction(
            error_details=FinishErrorDetails(
                error=MyCustomError(message="oh no"),
                error_id="error-id-1",
                created_at=datetime(year=2021, month=1, day=1),
            )
        )
    )

    subject.handle_action(
        HardwareStoppedAction(
            completed_at=datetime(year=2022, month=2, day=2),
            finish_error_details=FinishErrorDetails(
                error=MyCustomError(message="yikes"),
                error_id="error-id-2",
                created_at=datetime(year=2023, month=3, day=3),
            ),
        )
    )

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.FAILED,
        run_completed_at=datetime(year=2022, month=2, day=2),
        is_door_blocking=False,
        run_error=errors.ErrorOccurrence(
            id="error-id-1",
            createdAt=datetime(year=2021, month=1, day=1),
            errorType="MyCustomError",
            detail="oh no",
            errorCode=ErrorCodes.PIPETTE_NOT_PRESENT.value.code,
        ),
        finish_error=errors.ErrorOccurrence(
            id="error-id-2",
            createdAt=datetime(year=2023, month=3, day=3),
            errorType="MyCustomError",
            detail="yikes",
            errorCode=ErrorCodes.PIPETTE_NOT_PRESENT.value.code,
        ),
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=None,
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_ignores_stop_after_graceful_finish() -> None:
    """It should no-op on stop if already gracefully finished."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))
    subject.handle_action(FinishAction())
    subject.handle_action(StopAction())

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.SUCCEEDED,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=datetime(year=2021, month=1, day=1),
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_command_store_ignores_finish_after_non_graceful_stop() -> None:
    """It should no-op on finish if already ungracefully stopped."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )

    subject.handle_action(PlayAction(requested_at=datetime(year=2021, month=1, day=1)))
    subject.handle_action(StopAction())
    subject.handle_action(FinishAction())

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.STOPPED,
        run_completed_at=None,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=datetime(year=2021, month=1, day=1),
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()


def test_handles_hardware_stopped() -> None:
    """It should mark the hardware as stopped on HardwareStoppedAction."""
    subject = CommandStore(
        is_door_open=False,
        config=_make_config(),
        error_recovery_policy=_placeholder_error_recovery_policy,
    )
    completed_at = datetime(year=2021, day=1, month=1)
    subject.handle_action(
        HardwareStoppedAction(completed_at=completed_at, finish_error_details=None)
    )

    assert subject.state == CommandState(
        command_history=CommandHistory(),
        queue_status=QueueStatus.PAUSED,
        run_result=RunResult.STOPPED,
        run_completed_at=completed_at,
        is_door_blocking=False,
        run_error=None,
        finish_error=None,
        failed_command=None,
        command_error_recovery_types={},
        recovery_target=None,
        run_started_at=None,
        latest_protocol_command_hash=None,
        stopped_by_async_error=False,
        is_stopping_because_of_async_error=False,
        error_recovery_policy=matchers.Anything(),
        has_entered_error_recovery=False,
    )
    assert subject.state.command_history.get_running_command() is None
    assert subject.state.command_history.get_all_ids() == []
    assert subject.state.command_history.get_queue_ids() == OrderedSet()
    assert subject.state.command_history.get_setup_queue_ids() == OrderedSet()



"""Tests for the engine based Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import AbsorbanceReader
from opentrons.hardware_control.modules.types import (
    ModuleType,
)
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.engine.module_core import AbsorbanceReaderCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.errors.exceptions import CannotPerformModuleAction
from opentrons.protocol_engine.state.module_substates import AbsorbanceReaderSubState
from opentrons.protocol_engine.state.module_substates.absorbance_reader_substate import (
    AbsorbanceReaderId,
    AbsorbanceReaderMeasureMode,
)

SyncAbsorbanceReaderHardware = SynchronousAdapter[AbsorbanceReader]


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SyncAbsorbanceReaderHardware:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SyncAbsorbanceReaderHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def subject(
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncAbsorbanceReaderHardware,
    mock_protocol_core: ProtocolCore,
) -> AbsorbanceReaderCore:
    """Get a AbsorbanceReaderCore test subject."""
    return AbsorbanceReaderCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncAbsorbanceReaderHardware,
    mock_protocol_core: ProtocolCore,
) -> None:
    """It should be able to create an absorbance plate reader module core."""
    result = AbsorbanceReaderCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )

    assert result.module_id == "1234"
    assert result.MODULE_TYPE == ModuleType.ABSORBANCE_READER


def test_initialize(
    decoy: Decoy, mock_engine_client: EngineClient, subject: AbsorbanceReaderCore
) -> None:
    """It should set the sample wavelength with the engine client."""
    subject._ready_to_initialize = True
    subject.initialize("single", [350])

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.absorbance_reader.InitializeParams(
                moduleId="1234",
                measureMode="single",
                sampleWavelengths=[350],
                referenceWavelength=None,
            ),
        ),
        times=1,
    )
    assert subject._initialized_value == [350]

    # Test reference wavelength
    subject.initialize("single", [350], 450)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.absorbance_reader.InitializeParams(
                moduleId="1234",
                measureMode="single",
                sampleWavelengths=[350],
                referenceWavelength=450,
            ),
        ),
        times=1,
    )
    assert subject._initialized_value == [350]

    # Test initialize multi
    subject.initialize("multi", [350, 400, 450])

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.absorbance_reader.InitializeParams(
                moduleId="1234",
                measureMode="multi",
                sampleWavelengths=[350, 400, 450],
                referenceWavelength=None,
            ),
        ),
        times=1,
    )
    assert subject._initialized_value == [350, 400, 450]


def test_initialize_not_ready(subject: AbsorbanceReaderCore) -> None:
    """It should raise CannotPerformModuleAction if you dont call .close_lid() command."""
    subject._ready_to_initialize = False
    with pytest.raises(CannotPerformModuleAction):
        subject.initialize("single", [350])


@pytest.mark.parametrize("wavelength", [-350, 0, 1200, "wda"])
def test_invalid_wavelengths(wavelength: int, subject: AbsorbanceReaderCore) -> None:
    """It should raise ValueError if you provide an invalid wavelengthi."""
    subject._ready_to_initialize = True
    with pytest.raises(ValueError):
        subject.initialize("single", [wavelength])


def test_read(
    decoy: Decoy, mock_engine_client: EngineClient, subject: AbsorbanceReaderCore
) -> None:
    """It should call absorbance reader to read with the engine client."""
    subject._ready_to_initialize = True
    subject._initialized_value = [350]
    substate = AbsorbanceReaderSubState(
        module_id=AbsorbanceReaderId(subject.module_id),
        configured=True,
        measured=False,
        is_lid_on=True,
        data=None,
        configured_wavelengths=subject._initialized_value,
        measure_mode=AbsorbanceReaderMeasureMode("single"),
        reference_wavelength=None,
    )
    decoy.when(
        mock_engine_client.state.modules.get_absorbance_reader_substate(
            subject.module_id
        )
    ).then_return(substate)
    subject.read(filename=None)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.absorbance_reader.ReadAbsorbanceParams(
                moduleId="1234",
                fileName=None,
            ),
        ),
        times=1,
    )


def test_get_serial_number(
    decoy: Decoy, subject: AbsorbanceReaderCore, mock_engine_client: EngineClient
) -> None:
    """It should return a serial number."""
    decoy.when(mock_engine_client.state.modules.get_serial_number("1234")).then_return(
        "abc"
    )

    assert subject.get_serial_number() == "abc"
